@@ -20,11 +20,7 @@ import os
 import re
 import smtplib
 import types
-
-try:
-    import email.Utils as Utils
-except ImportError:
-    import email.utils as Utils
+import email.utils as Utils
 
 import pwd
 
@@ -36,13 +32,13 @@ __all__ = [
     'DeliverySkeleton',
     'Maildir',
     'Mboxrd',
-    'MDA_qmaillocal',
     'MDA_external',
     'MDA_lmtp',
-    'MultiDestinationBase',
+    'MDA_qmaillocal',
     'MultiDestination',
-    'MultiSorterBase',
+    'MultiDestinationBase',
     'MultiSorter',
+    'MultiSorterBase',
 ]
 
 def exit_with(stderr,errmsg):
@@ -200,10 +196,13 @@ class Maildir(DeliverySkeleton, ForkingBase):
                         'refuse to deliver mail as GID 0'
                     )
 
-        child = self.forkchild(
-            lambda o,e: self.__deliver_message_maildir(
-                uid, gid, msg, delivered_to, received, o, e)
-            )
+        try:
+            child = self.forkchild(
+                lambda o,e: self.__deliver_message_maildir(
+                    uid, gid, msg, delivered_to, received, o, e)
+                )
+        except getmailOperationError as oe:
+            raise getmailDeliveryError( 'delivery %s child operation error: %s' % (self, oe) )
 
         self.log.debug('maildir delivery process %d exited %d\n'
                        % (child.childpid, child.exitcode))
@@ -275,10 +274,7 @@ class Mboxrd(DeliverySkeleton, ForkingBase):
             # Open mbox file, refusing to create it if it doesn't exist
             fd = os.open(self.conf['path'], os.O_RDWR)
             status_old = os.fstat(fd)
-            try:
-                f = os.fdopen(fd, 'br+')
-            except ValueError: # Py2
-                f = os.fdopen(fd, 'r+')
+            f = os.fdopen(fd, 'br+')
             lock_file(f, self.conf['locktype'])
             # Check if it _is_ an mbox file.  mbox files must start with "From "
             # in their first line, or are 0-length files.
@@ -355,10 +351,13 @@ class Mboxrd(DeliverySkeleton, ForkingBase):
                     'refuse to deliver mail as GID 0'
                 )
 
-        child = self.forkchild(
-            lambda o,e: self.__deliver_message_mbox(
-                uid, gid, msg, delivered_to, received, o, e)
-            )
+        try:
+            child = self.forkchild(
+                lambda o,e: self.__deliver_message_mbox(
+                    uid, gid, msg, delivered_to, received, o, e)
+                )
+        except getmailOperationError as oe:
+            raise getmailDeliveryError( 'delivery %s child operation error: %s' % (self, oe) )
 
         self.log.debug('mboxrd delivery process %d exited %d\n'
                        % (child.childpid, child.exitcode))
@@ -462,7 +461,7 @@ class MDA_qmaillocal(DeliverySkeleton, ForkingBase):
     def showconf(self):
         self.log.info('MDA_qmaillocal(%s)\n' % self._confstring())
 
-    def _deliver_qmaillocal(self, msg, msginfo, delivered_to, received, stdout,
+    def __deliver_qmaillocal(self, msg, msginfo, delivered_to, received, stdout,
             stderr):
         try:
             args = (
@@ -522,10 +521,13 @@ class MDA_qmaillocal(DeliverySkeleton, ForkingBase):
         self.log.debug('recipient: set dash to "%s", ext to "%s"\n'
                        % (msginfo['dash'], msginfo['ext']))
 
-        child = self.forkchild(
-            lambda o,e: self._deliver_qmaillocal(
-                msg, msginfo, delivered_to, received, o, e)
-            )
+        try:
+            child = self.forkchild(
+                lambda o,e: self.__deliver_qmaillocal(
+                    msg, msginfo, delivered_to, received, o, e)
+                )
+        except getmailOperationError as oe:
+            raise getmailDeliveryError( 'delivery %s child operation error: %s' % (self, oe) )
 
 
         self.log.debug('qmail-local %d exited %d\n' % (child.childpid, child.exitcode))
@@ -592,6 +594,9 @@ class MDA_external(DeliverySkeleton, ForkingBase):
 
       ignore_stderr (boolean, optional) - if set, getmail will not consider the
             program writing to stderr to be an error.  The default is False.
+
+      pipe_stdout (boolean, optional) - if set, stdout from external command
+            will be forwarded to stdout of getmail.  The default is True.
     '''
     _confitems = (
         ConfInstance(name='configparser', required=False),
@@ -602,6 +607,7 @@ class MDA_external(DeliverySkeleton, ForkingBase):
         ConfBool(name='allow_root_commands', required=False, default=False),
         ConfBool(name='unixfrom', required=False, default=False),
         ConfBool(name='ignore_stderr', required=False, default=False),
+        ConfBool(name='pipe_stdout', required=False, default=True),
     )
 
     def initialize(self):
@@ -610,7 +616,7 @@ class MDA_external(DeliverySkeleton, ForkingBase):
         if not os.access(self.conf['path'], os.X_OK):
             raise getmailConfigurationError('%s not executable'
                                             % self.conf['path'])
-        if type(self.conf['arguments']) != tuple:
+        if not isinstance(self.conf['arguments'], tuple):
             raise getmailConfigurationError(
                 'incorrect arguments format; see documentation (%s)'
                 % self.conf['arguments']
@@ -624,13 +630,11 @@ class MDA_external(DeliverySkeleton, ForkingBase):
     def showconf(self):
         self.log.info('MDA_external(%s)\n' % self._confstring())
 
-    def _deliver_command(self, msg, msginfo, delivered_to, received,
+    def __deliver_command(self, msg, msginfo, delivered_to, received,
                          stdout, stderr):
         try:
             args = [self.conf['path'], self.conf['path']]
             msginfo['mailbox'] = self.retriever.mailbox_selected or ''
-            if sys.version_info.major == 2:
-                msginfo['mailbox'] = msginfo['mailbox'].encode('utf-8')
             for arg in self.conf['arguments']:
                 arg = expand_user_vars(arg)
                 for (key, value) in msginfo.items():
@@ -650,10 +654,16 @@ class MDA_external(DeliverySkeleton, ForkingBase):
         self.log.trace()
         msginfo = self.get_msginfo(msg)
 
-        child = self.forkchild(
-            lambda o,e: self._deliver_command(
-                msg, msginfo, delivered_to, received, o, e)
-            )
+        try:
+            child = self.forkchild(
+                lambda o,e: self.__deliver_command(
+                    msg, msginfo, delivered_to, received, o, e)
+                )
+        except getmailOperationError as oe:
+            raise getmailDeliveryError( 'delivery %s child operation error: %s' % (self, oe) )
+
+        if self.conf['pipe_stdout'] and child.out:
+            self.log.info('%s\n' % (''.join(map(chr, child.out))))
 
         self.log.debug('command %s %d exited %d\n'
                        % (self.conf['command'], child.childpid, child.exitcode))
@@ -685,10 +695,6 @@ class MDA_lmtp(DeliverySkeleton):
     )
 
     def initialize(self):
-        if sys.version_info.major < 3:
-            raise getmailConfigurationError(
-                'MDA_lmtp in config: that was implemented only for python 3'
-            )
         self.log.trace()
         self.__connect()
 
@@ -733,7 +739,7 @@ class MDA_lmtp(DeliverySkeleton):
 
     def _deliver_message(self, msg, delivered_to, received):
         self.log.trace()
-        recipient = self.conf.get('override', msg.recipient)
+        recipient = self.conf.get('override', msg.recipient) or None
         rcpt = self.__send(msg.content(), msg.sender, recipient)
 
         if recipient in rcpt:
@@ -893,12 +899,12 @@ class MultiSorterBase(MultiDestinationBase):
         try:
             _locals = self.conf['locals']
             # Special case for convenience if user supplied one base 2-tuple
-            if (len(_locals) == 2 and type(_locals[0]) == str
-                    and type(_locals[1]) == str):
+            if (len(_locals) == 2 and
+                isinstance(_locals[0], str) and isinstance(_locals[1], str)):
                 _locals = (_locals, )
             for item in _locals:
-                if not (type(item) == tuple and len(item) == 2
-                        and type(item[0]) == str and type(item[1]) == str):
+                if (not isinstance(item, tuple) and len(item) == 2 and
+                    isinstance(item[0], str) and isinstance(item[1], str)):
                     raise getmailConfigurationError(
                         'invalid syntax for locals; see documentation'
                     )
@@ -956,7 +962,7 @@ class MultiSorter(MultiSorterBase):
                ("jason@example.org", "/home/jasonk/Maildir/"),
                ("sales@example.org", "/home/karlyk/Mail/sales"),
                ("abuse@(example.org|example.net)", "/home/kellyw/Mail/abuse/"),
-               ("^(jeff|jefferey)(\.s(mith)?)?@.*$", "[jeff-mail-delivery]"),
+               ("^(jeff|jefferey)(\\.s(mith)?)?@.*$", "[jeff-mail-delivery]"),
                ("^.*@(mail.)?rapinder.example.org$", "/home/rapinder/Maildir/")
                )
 
